/*\

Copyright 2017 The University of North Carolina at Chapel Hill.
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation for educational, research and non-profit purposes, without
fee, and without a written agreement is hereby granted, provided that the
above copyright notice and the following three paragraphs appear in all
copies.

IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
DAMAGES.

THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The authors may be contacted via:

US Mail: Real-Time Systems Group at UNC
Department of Computer Science
Sitterson Hall
University of N. Carolina
Chapel Hill, NC 27599-3175

EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu

**/

#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "fast_rwrnlp.h"

/*
 *  Fast RW-RNLP: initialize
 */
void fast_rwrnlp_init(fast_rwrnlp_t *lock)
{
    unsigned i;

    // Initialize the RW-RNLP*
    rwrnlp_star_init(&lock->rwrnlp_star);

    // Initialize all ticket locks (used for non-nested writes)
    for (i = 0; i < NR_RESOURCES; i++)
    {
        ticket_init(&lock->ticketlocks[i]);
    }

    // Initialize the RNLP, and zero out per-processor nodes
    // (used for nested writes)
    rnlp_init(&lock->rnlp);
    for (i = 0; i < NR_CPUS; i++)
    {
        lock->rnlpnodes[i].requested = 0;
        lock->rnlpnodes[i].blocked = 0;
        lock->rnlpnodes[i].next = NULL;
    }

    // Initialize the ticket lock for accessing data structures
    lock->nested = (ticket_lock_t *)(malloc(sizeof(ticket_lock_t)));
    ticket_init(lock->nested);
}

/*
 *  Non-nested read: lock
 */
void nn_read_lock(fast_rwrnlp_t *lock, int resource)
{
    rwrnlp_star_nn_read_lock(&lock->rwrnlp_star, resource);
}

/*
 *  Non-nested: unlock
 */
void nn_read_unlock(fast_rwrnlp_t *lock, int resource)
{
    rwrnlp_star_nn_read_unlock(&lock->rwrnlp_star, resource);
}

/*
*  Nested read: lock
*/
void n_read_lock(fast_rwrnlp_t *lock, resource_mask_t resources, int processor)
{
    rwrnlp_star_read_lock(&lock->rwrnlp_star, resources, processor);
}

/*
*  Nested read: unlock
*/
void n_read_unlock(fast_rwrnlp_t *lock, resource_mask_t resources, int processor)
{
    rwrnlp_star_read_unlock(&lock->rwrnlp_star, resources, processor);
}

/*
 *  Non-nested write: lock (use RW-RNLP*)
 */
void nn_write_lock(fast_rwrnlp_t *lock, int resource)
{
    // First wait to get past the nn-write ticket lock for this resource
    ticket_lock(&lock->ticketlocks[resource]);

    rwrnlp_star_nn_write_lock(&lock->rwrnlp_star, resource);

}

/*
 *  Non-nested write: unlock (use RW-RNLP*)
 */
void nn_write_unlock(fast_rwrnlp_t *lock, int resource)
{
    // Note: only one write for this resource (and thus this lock) can be
    // here at a time because all other writes are blocked on the wticket
    // wait in s_write_lock

    rwrnlp_star_nn_write_unlock(&lock->rwrnlp_star, resource);

    ticket_unlock(&lock->ticketlocks[resource]);
}

/*
 *  Nested write: lock (use RW-RNLP*)
 */
void n_write_lock(fast_rwrnlp_t *lock, resource_mask_t resources, int processor)
{
    rnlp_node_t *rnlp_node;

    // Look up the RNLP node for this processor and update its resources
    rnlp_node = &lock->rnlpnodes[processor];
    rnlp_node->requested = resources;
    rnlp_node->blocked = 0;
    rnlp_node->next = NULL;

    // First wait to get past the n-write RNLP
    rnlp_lock(&lock->rnlp, rnlp_node);

    // Now call rwrnlp_star
    rwrnlp_star_write_lock(&lock->rwrnlp_star, resources, processor);
}

/*
 *  Nested write: unlock (use RW-RNLP*)
 */
void n_write_unlock(fast_rwrnlp_t *lock, resource_mask_t resources, int processor)
{
    rnlp_node_t *rnlp_node;

    rwrnlp_star_write_unlock(&lock->rwrnlp_star, resources, processor);

    // Look up the RNLP node for this processor
    rnlp_node = &lock->rnlpnodes[processor];

    // Release the n-write RNLP
    rnlp_unlock(&lock->rnlp, rnlp_node);
}
